import datetime
import re


class datedelta:
    __slots__ = '_months', '_timedelta'

    def __new__(cls, months=0, days=0, seconds=0, microseconds=0,
                milliseconds=0, minutes=0, hours=0, weeks=0, years=0):

        self = object.__new__(cls)
        self._timedelta = datetime.timedelta(days, seconds, microseconds, milliseconds, minutes, hours, weeks)
        if not isinstance(months, int):
            raise TypeError('months must be an integer')
        self._months = months + 12 * years
        return self

    @classmethod
    def from_timedelta(cls, td):
        return datedelta(0, 0, td.days, td.seconds, td.microseconds)

    @classmethod
    def from_natural_str(cls, s):
        m = re.match("((?P<years>[\d]*)[\s]*y(ear(s)*)*)*(,\s)*((?P<months>[\d]*)[\s]*m(onth(s)*)*)*(,\s)*((?P<days>[\d]*)[\s]*d(ay(s)*)*)*(,\s)*", s)
        if m is None:
            raise ValueError("cannot match string format (e.g: '1m, 1 days')")
        years = int(m.group("years")) if m.group("years") else 0
        months = int(m.group("months")) if m.group("months") else 0
        months+= years * 12
        days = int(m.group("days")) if m.group("days") else 0
        return datedelta(months, days)

    def __repr__(self):
        args = []
        if self._months:
            args.append("months=%d" % self._months)
        if self._timedelta.days:
            args.append("days=%d" % self._timedelta.days)
        if self._timedelta.seconds:
            args.append("seconds=%d" % self._timedelta.seconds)
        if self._timedelta.microseconds:
            args.append("microseconds=%d" % self._timedelta.microseconds)
        if not args:
            args.append('0')
        return "%s.%s(%s)" % (self.__class__.__module__,
                              self.__class__.__qualname__,
                              ', '.join(args))

    def __str__(self):
        mm, ss = divmod(self._timedelta.seconds, 60)
        hh, mm = divmod(mm, 60)
        s = "%d:%02d:%02d" % (hh, mm, ss)

        def plural(n):
            return n, abs(n) != 1 and "s" or ""

        if self._timedelta.days:
            s = ("%d day%s, " % plural(self._timedelta.days)) + s
        if self._months:
            s = ("%d month%s, " % plural(self._months)) + s
        if self._timedelta.microseconds:
            s = s + ".%06d" % self._timedelta.microseconds
        return s

    # Read-only field accessors
    @property
    def months(self):
        """months"""
        return self._months

    @property
    def days(self):
        """days"""
        return self._timedelta.days

    def __add__(self, other):
        if isinstance(other, datetime.timedelta):
            other = datedelta.from_timedelta(other)
        if isinstance(other, datedelta):
            # for CPython compatibility, we cannot use
            # our __class__ here, but need a real timedelta
            return datedelta(self._months + other._months,
                             self._days + other._days,
                             self._seconds + other._seconds,
                             self._microseconds + other._microseconds)
        if isinstance(other, datetime.datetime):
            other += self._timedelta
            if self._months > 0:
                year = self._months//12 + other.year
                if self._months%12 + other.month > 12:
                    year += 1 
                    month = self._months%12 - (12-other.month)
                else:
                    month = self._months%12 + other.month
                return other.replace(year=year, month=month)
            elif self._months < 0:
                year = other.year - abs(self._months)//12
                if other.month - abs(self._months)%12 < 1:
                    year -= 1 
                    month = 12 - (abs(self._months)%12 - other.month)
                else:
                    month = other.month - abs(self._months)%12
                return other.replace(year=year, month=month)
            return other
        return NotImplemented

    __radd__ = __add__

    def __sub__(self, other):
        return -self + other

    def __rsub__(self, other):
        if isinstance(other, datetime.datetime):
            return -self + other
        return NotImplemented

    def __neg__(self):
        n = datedelta(months=-self._months)
        n._timedelta = -self._timedelta
        return n

    def __pos__(self):
        return self

    def __abs__(self):
        if self._months < 0 and self._timedelta<0:
            return -self
        else:
            return self

    def __mul__(self, other):
        if isinstance(other, int):
            return datedelta(self._months * other,
                             self._timedelta.days * other,
                             self._timedelta.seconds * other,
                             self._timedelta.microseconds * other)
        return NotImplemented

    __rmul__ = __mul__


import logging
import requests
from requests_toolbelt.utils import dump
import base64

log = logging.getLogger("enedis-login")

class LoginException(Exception):
    pass

class Login:
    URL = 'https://espace-client-connexion.enedis.fr/auth/UI/Login'

    def __init__(self, url=None):
        self.url = url if url is not None else Login.URL
        self.session = requests.Session()
        log.debug("Creating login (%s) with session %s" % (self.url, self.session))

    def login(self, username, password):
        self.session.headers.update({'User-agent': "mylinky"})
        payload = {
            'IDToken1': username,
            'IDToken2': password,
            'SunQueryParamsString': base64.b64encode(b'realm=particuliers'),
            'encoded': 'true',
            'gx_charset': 'UTF-8'
        }

        log.info("Sending login request for user %s" % username)
        resp = self.session.post(self.url, data=payload, allow_redirects=False)

        log.debug("resp: %s" % dump.dump_response(resp).decode('utf-8'))
        log.debug("cookies: %s" % resp.cookies)
        if not 'iPlanetDirectoryPro' in resp.cookies:
            raise LoginException("Login unsuccessful. Check your credentials.")

        return [ c for c in resp.cookies]



import logging
import datetime

from .login import Login
from .data import Data

log = logging.getLogger("enedis")

class Enedis:
    RESOURCE = {
        "hourly": Data.RESOURCE_HOURLY,
        "monthly": Data.RESOURCE_MONTHLY,
        "yearly": Data.RESOURCE_YEARLY
    }

    def __init__(self, url=None, url2=None, timesheets=None):
        self._login = Login(url)
        self._url2 = url2
        self._cookies = None
        self._timesheets = timesheets

    def login(self, username, password):
        cookies = self._login.login(username, password)
        self._cookies = cookies

    @classmethod
    def parsetimesheet(cls, start, end):
        return (datetime.datetime.strptime(start, "%H:%M").time(), datetime.datetime.strptime(end, "%H:%M").time())

    def getdata(self, kind, **kwargs):
        h = Data(self._cookies, url=self._url2, timesheets=self._timesheets)

        # normalize date
        if kind == "monthly":
            kwargs["startDate"] = kwargs["startDate"].replace(day=1, hour=0, minute=0, second=0, microsecond=0)
            kwargs["endDate"] = kwargs["endDate"].replace(day=1, hour=0, minute=0, second=0, microsecond=0)
        if kind == "yearly":
            kwargs["startDate"] = kwargs["startDate"].replace(day=1, month=1, hour=0, minute=0, second=0, microsecond=0)
            kwargs["endDate"] = kwargs["endDate"].replace(day=1, month=1, hour=0, minute=0, second=0, microsecond=0)

        log.debug("requesting data (%s): %s - %s" % (kind, kwargs["startDate"], kwargs["endDate"]))
        data = h.get_data(resource=Enedis.RESOURCE[kind], **kwargs)
        log.debug("data (%d): %s%s" % (len(data), data[0:5], "..." if len(data)>6 else ""))
        return data




#!/usr/local/bin/python3
# encoding: utf-8
'''
linky -- Toolbox for dumping data from your French LINKY power counter (from ENEDIS)

It defines classes_and_methods

@author:     lmolinier

@license:    MIT

@contact:    lionel@molinier.eu
@deffield    updated: Updated
'''

import sys
import os
import logging
import pkg_resources
import datetime
import argparse
import pprint

from argparse import ArgumentParser
from argparse import RawDescriptionHelpFormatter

from mylinky.datedelta import datedelta
from mylinky.enedis import Enedis
from mylinky.exporter import InfluxdbExporter
from mylinky import MyLinkyConfig

__all__ = []
__version__ = "UNKNOWN"
__date__ = '2019-11-15'

try:
    __version__ = pkg_resources.get_distribution("mylinky").version
except pkg_resources.DistributionNotFound:
    pass

# Default logging configuration
logging.basicConfig(format='%(asctime)s %(message)s') #, datefmt='%m/%d/%Y %I:%M:%S %p')
log = logging.getLogger()
log.setLevel(logging.INFO)

DEBUG = 1
TESTRUN = 0
PROFILE = 0

def datetime_converter(s):
    try:
        return datetime.datetime.strptime(s, "%d/%m/%Y")
    except ValueError:
        raise argparse.ArgumentTypeError("Invalid date (DD/MM/YYYY): %s" % s)

def datedelta_converter(s):
    try:
        return datedelta.from_natural_str(s)
    except ValueError:
        raise argparse.ArgumentTypeError("Cannot parse date delta (e.g: '1d', '1m', ...)")

def timesheet_converter(s):
    try:
        (begin,end) = s.split(",")
        return Enedis.parsetimesheet(begin, end)
    except ValueError:
        raise argparse.ArgumentTypeError("Cannot parse timesheet 'time:time', with time '%h:%m'")

def main(argv=None): # IGNORE:C0111
    '''Command line options.'''
    if argv is None:
        argv = sys.argv
    else:
        sys.argv.extend(argv)

    program_name = os.path.basename(sys.argv[0])
    program_version_message = '%%(prog)s %s' % (__version__)
    program_shortdesc = "LINKY Power consumption toolbox (from ENEDIS)"
    program_license = '''%s

  Created by lmolinier on %s.

  Distributed on an "AS IS" basis without warranties
  or conditions of any kind, either express or implied.

USAGE
''' % (program_shortdesc, str(__date__))

    try:
        config = MyLinkyConfig()

        # Setup argument parser
        parser = ArgumentParser(description=program_license, formatter_class=RawDescriptionHelpFormatter)
        parser.add_argument('-v', '--verbose', action='count', default=0, help="set verbosity level [default: %(default)s]")
        parser.add_argument('-V', '--version', action='version', version=program_version_message)
        parser.add_argument('-c', "--config", help="configuration file")
        
        group = parser.add_argument_group("importer", "Data importer parameters")
        group.add_argument("--importer", choices=["enedis"], help="select importer [default: %(default)s]", default="enedis")

        enedis = group.add_argument_group("enedis", "ENEDIS Access and Data paremeters")
        enedis.add_argument('-u', '--username', help="Enedis username")
        enedis.add_argument('-p', '--password', help="Enedis password")
        enedis.add_argument("--timesheet", action="append", type=timesheet_converter, help="enter new HP/HC timesheet")

        parser.add_argument("--type", choices=Enedis.RESOURCE.keys(), default="hourly", help="query data source (default: %(default)s)")

        date = parser.add_argument_group("date range", "select the date range")
        date.add_argument("--to", help="to/end query data range (format DD/MM/YYYY)", type=datetime_converter, default=datetime.datetime.now())
        group = date.add_mutually_exclusive_group()
        group.add_argument("--from", help="from/start query date range (format DD/MM/YYYY)", type=datetime_converter)
        group.add_argument("--last", help="query for last days/months/year depending 'type'", type=datedelta_converter)

        subparsers = parser.add_subparsers(help='exporter help', dest="exporter")

        subparser = subparsers.add_parser("influxdb", help="Export to InfluxDB")
        subparser.add_argument("--host", help="Database hostname")
        subparser.add_argument("--db", help="Database name")
        subparser.add_argument("--dbuser", help="Database username")
        subparser.add_argument("--dbpassword", help="Database password")

        subparser = subparsers.add_parser("stdout", help="Export to STDOUT")
        subparser.add_argument("--pretty", help="enable pretty printing")

        args = parser.parse_args()
        kwargs = vars(args)

        log.info('setting log level to {}'.format(args.verbose))
        levels = [logging.CRITICAL, logging.ERROR, logging.WARNING, logging.INFO, logging.DEBUG]
        level = levels[max(0, min(args.verbose-1, len(levels)-1))]
        log.info('effective verbose value set to {}'.format(level))
        logging.getLogger().setLevel(level)

        log.debug("args: %s" % args)
        log.debug("kwargs: %s" % kwargs)
        if args.config:
            config.load_from_file(args.config)
        config.override_from_args(kwargs)
        log.debug("config: %s" % config.data)

        enedis = Enedis(timesheets=config["enedis"]["timesheets"])
        enedis.login(config["enedis"]["username"], config["enedis"]["password"])

        startDate = kwargs["from"]
        endDate = kwargs["to"]
        if startDate is None:
            delta = args.last
            if not delta:
                delta = {"hourly": datedelta(days=1), "monthly": datedelta(months=1), "yearly": datedelta(years=1) }[args.type]
            startDate = endDate - delta

        data = enedis.getdata(args.type, startDate=startDate, endDate=endDate)

        if args.exporter == "influxdb":
            influx = InfluxdbExporter(
                host=config["influxdb"]["host"], 
                port=config["influxdb"]["port"],
                database=config["influxdb"]["database"],
                username=config["influxdb"]["username"],
                password=config["influxdb"]["password"],
                prefix=config["influxdb"]["measurement-prefix"]
            )

            influx.save_data(args.type, data)
        elif args.exporter == "stdout":
            pprint.pprint(data)

        return 0

    except Exception as e:
        if DEBUG or TESTRUN:
            raise(e)
        indent = len(program_name) * " "
        sys.stderr.write(program_name + ": " + repr(e) + "\n")
        sys.stderr.write(indent + "  for help use --help")
        return 2

    return 0

if __name__ == "__main__":
    if DEBUG:
        #sys.argv.append("-h")
        #sys.argv.insert(1, "-v")
        #sys.argv.append("-r")
        pass
    if TESTRUN:
        import doctest
        doctest.testmod()
    if PROFILE:
        import cProfile
        import pstats
        profile_filename = 'profile.txt'
        cProfile.run('main()', profile_filename)
        statsfile = open("profile_stats.txt", "wb")
        p = pstats.Stats(profile_filename, stream=statsfile)
        stats = p.strip_dirs().sort_stats('cumulative')
        stats.print_stats()
        statsfile.close()
        sys.exit(0)
    main()



from influxdb import InfluxDBClient

class InfluxdbExporter:
    def __init__(self, **kwargs):
        self.prefix = kwargs["prefix"]
        del kwargs["prefix"]
        self.client = InfluxDBClient(**kwargs)

    def connect(self):
        pass

    def save_data(self, serie, data, batchid=0):
        d = []
        for item in data:
            d.append({
                "measurement": "%s%s"% (self.prefix, serie),
                "tags": {
                    "batchid" : batchid
                },
                "time": item['date'].strftime('%Y-%m-%dT%H:%M:%SZ'),
                "fields": {
                    "value": (item['value']*1000),
                    "duration": item['duration']
                }
            })

        self.client.write_points(d)
        return len(d)


import logging
import requests
import datetime
from requests_toolbelt.utils import dump
import base64

from mylinky.datedelta import datedelta

log = logging.getLogger("enedis-data")

class DataException(Exception):
    pass

class Data:
    URL = 'https://espace-client-particuliers.enedis.fr/group/espace-particuliers/suivi-de-consommation'

    RESOURCE_HOURLY = "urlCdcHeure"
    RESOURCE_MONTHLY = "urlCdcMois"
    RESOURCE_YEARLY = "urlCdcAn"

    STEPS = {
        RESOURCE_HOURLY: datedelta(minutes=30),
        RESOURCE_MONTHLY: datedelta(months=1),
        RESOURCE_YEARLY: datedelta(years=1),
    }

    def __init__(self, cookies, timesheets=None, url=None):
        self.url = url if url is not None else Data.URL
        self.session = requests.Session()
        self.timesheets = timesheets

        if cookies is not None:
            if type(cookies) != list:
                cookies = [cookies]
            for cookie in cookies:
                self.session.cookies.set_cookie(cookie)

        log.debug("Creating data (%s) with session %s" % (self.url, self.session))

    def get_data(self, resource, startDate=datetime.datetime(year=1970, month=1, day=1), endDate=datetime.datetime.today()):
        raw = self._query_data(resource, startDate, endDate)

        data = self._transform_data(resource, raw, (startDate, endDate))
        return data

    def _get_type(self, startdate):
        daytime = startdate.time()
        if self.timesheets is None:
            return "normale"
        
        for (stime, etime) in self.timesheets:
            if stime<etime:
                # not crossing a day, this is
                # the normal case
                # e.g.: 00:00 --> 06:00
                if daytime >= stime and daytime < etime:
                    return "creuse"
            else:
                # we are crossing a day
                # e.g.: 22:00 --> 06:30
                if daytime >= stime or daytime < etime:
                    return "creuse"

        return "pleine"

        

    def _transform_data(self, resource, raw, bounds=None):
        start = datetime.datetime.strptime(raw["periode"]["dateDebut"], "%d/%m/%Y")
        end = datetime.datetime.strptime(raw["periode"]["dateFin"], "%d/%m/%Y")
        step = Data.STEPS[resource]

        if resource == Data.RESOURCE_YEARLY:
            start = start.replace(day=1, month=1)
            end = end.replace(day=1, month=1)

        data = []
        offset = raw["decalage"] - 1 if raw["decalage"]>0 else 0
        for item in raw["data"]:
            rank = int(item["ordre"])-1

            # apparently, the yearly data is a bit different and start its 'ordre' to 0, instead of 1...
            if resource == Data.RESOURCE_YEARLY:
                rank+=1

            # correct the start with the 'decalage' field for incomplete graphe
            if rank < offset:
                continue
            rank = rank - offset
            value = float(item["valeur"])

            begin = start + (rank*step)
            end = start + ((rank+1)*step)
            duration = end - begin

            if bounds and (begin < bounds[0] or begin >= bounds[1]):
                continue

            # Value is given in kW
            #  - if value is '-2', there is no value --> drop
            #  - if value is '-1', TODO
            if item["valeur"] < 0:
                continue
            d = {'date': begin, 'duration': duration.total_seconds(), 'value': value}
            if resource == Data.RESOURCE_HOURLY:
                d.update({'type': self._get_type(start + (rank*step))})
            data.append(d)

        return data

    def _query_data(self, resource, startDate, endDate):

        self.session.headers.update({'User-agent': "mylinky"})

        # note: payload is useless for yearly resource
        payload = {
            '_lincspartdisplaycdc_WAR_lincspartcdcportlet_dateDebut': startDate.strftime("%d/%m/%Y"),
            '_lincspartdisplaycdc_WAR_lincspartcdcportlet_dateFin': endDate.strftime("%d/%m/%Y")
        }

        params = {
            'p_p_id': 'lincspartdisplaycdc_WAR_lincspartcdcportlet',
            'p_p_lifecycle': 2,
            'p_p_state': 'normal',
            'p_p_mode': 'view',
            'p_p_resource_id': resource,
            'p_p_cacheability': 'cacheLevelPage',
            'p_p_col_id': 'column-1',
            'p_p_col_pos': 1,
            'p_p_col_count': 3
        }

        log.info("Sending data request for resource %s (%s - %s)" % (resource, startDate, endDate))
        resp = self.session.post(self.url, data=payload, params=params, allow_redirects=False)

        if 300 <= resp.status_code < 400:
            # It appears that it is frequent to get first a 302, even if the request is correct
            # #nocomment
            resp = self.session.post(self.url, data=payload, params=params, allow_redirects=False)

        log.debug("resp: %s" % dump.dump_response(resp).decode('utf-8'))
        body = resp.json()

        if body["etat"]["valeur"] == "erreur":
            raise DataException("Error on server when retrieving data: %s" % body["etat"]["erreurText"] if "erreurText" in body["etat"] else "n/a")

        if body["etat"]["valeur"] not in ["termine"]:
            raise DataException("Invalid response state code '%s'" % body["etat"]["valeur"])

        return body["graphe"]




import yaml
import logging

from mylinky.enedis import Enedis

log = logging.getLogger("config")


class MyLinkyConfig():
    def __init__(self):
        self.data = { 
            "enedis": {
                "username": None,
                "password": None,
                "timesheets": None
            },
            "influxdb" : {
                "host": "localhost",
                "port": 8086,
                "database": "linky",
                "username": None,
                "password": None,
                "measurement-prefix": "linky_"
            }
        }

    def _merge(self, d2, d=None):
        if d is None:
            d = self.data
        for (k,v) in d.items():
            if type(d[k])==dict:
                self._merge(d2[k] if k in d2 else {}, d[k])
            elif k in d2:
                if k=="timesheets":
                    d[k] = [ Enedis.parsetimesheet(l[0],l[1]) for l in d2[k] ]
                else:
                    d[k] = d2[k]
        return d

    def load_from_file(self, fname):
        with open(fname) as f:
            data = yaml.load(f)
        log.debug("using configuration: %s" % data)
        self._merge(data)
        log.debug("config: %s" % self.data)
        return self

    def __getitem__(self, key):
        return self.data[key]

    def override_from_args(self, kwargs):
        # ENEDIS Access
        if "username" in kwargs and kwargs["username"] is not None:
            self.data["enedis"]["user"] = kwargs["username"]
        if "password" in kwargs and kwargs["password"] is not None:
            self.data["enedis"]["password"] = kwargs["password"]
        if "timesheet" in kwargs and kwargs["timesheet"] is not None:
            self.data["enedis"]["timesheets"] = kwargs["timesheet"]

        ## INFLUXDB
        if "host" in kwargs and kwargs["host"] is not None:
            (host,port) = kwargs["host"].split(":")
            self.data["influxdb"]["host"] = host
            self.data["influxdb"]["port"] = port
        if "dbuser" in kwargs and kwargs["dbuser"] is not None:
            self.data["influxdb"]["username"] = kwargs["dbuser"]
        if "dbpassword" in kwargs and kwargs["dbpassword"] is not None:
            self.data["influxdb"]["password"] = kwargs["dbpassword"]
        if "db" in kwargs and kwargs["db"] is not None:
            self.data["influxdb"]["database"] = kwargs["db"]




import subprocess

from setuptools import setup, find_packages

with open('requirements.txt') as f:
    required = f.read().splitlines()

with open("README.md") as f:
    readme = f.read()

def get_git_version(abbrev=7):
    def is_dirty():
        ## Check for dirty version
        try:
            p = subprocess.Popen(["git", "diff-index", "--name-only", "HEAD"],
                    stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            p.stderr.close()
            lines = p.stdout.readlines()
            return len(lines) > 0
        except:
            return False

    ## Call 'git describe --abbrev'
    try:
        p = subprocess.Popen(['git', 'describe', '--tags', '--abbrev=%d' % abbrev],
                  stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        p.stderr.close()
        line = p.stdout.readlines()[0].decode('utf-8')

        version = line.strip()
        if is_dirty():
            version+= "-dirty"
        return version

    except Exception as e:
        print(e)
        return None

setup(
    name='mylinky',
    version=get_git_version(),
    url='https://github.com/x3n1x/linky',
    license='MIT',
    author='Lionel Molinier',
    author_email='lionel@molinier.eu',
    description='Linky utility to grab your power consumption from ENEDIS',
    long_description = readme,
    long_description_content_type = "text/markdown",

    install_requires=required,
    packages=find_packages(exclude=["*.tests", "*.tests.*", "tests.*"]),

    entry_points = {
		"console_scripts": [
			"mylinky = mylinky.entrypoint:main"
		]
	},
)
